package se.chalmers.tdv.hashtable;

import java.util.LinkedList;
import java.util.Random;

import se.chalmers.tdv.hashtable.LLHashTableAction.ActionType;
import se.chalmers.tdv.testing.StatefulTest;
import se.chalmers.tdv.testing.StatefulTestCase;

/**
 * Implements the stateful test case for hash tables. In addition, adds
 * functionality to minimize the capacity of a failing test case and to print
 * the code that forms a failing test case.
 */
public class LLHashTableTestCase extends
		StatefulTestCase<LLHashTable, LLHashTableAction, Integer> {

	/** Maximum size of the hash table under test. */
	public static final int MAX_NRBUCKETS = 5;

	/**
	 * Maximum value of the keys in the hash table. Recommended to keep low to
	 * get more interesting random behavior.
	 */
	public static final int MAX_KEY = 15;

	/** The capacity of the hash tables in this test case. */
	private int nrBuckets;

	public LLHashTableTestCase(
			StatefulTest<LLHashTable, LLHashTableAction, Integer> test) {
		super(test);
	}

	@Override
	public void perform(LLHashTable t, LLHashTableAction a,
			LinkedList<Integer> outputArray) {
		switch (a.type) {
		case PUT:
			t.put(a.key, a.value);
			break;
		case REMOVE:
			t.remove(a.key);
			break;
		case GET:
			outputArray.add(t.get(a.key));
			break;
		case OUTPUT:
			outputArray.add(a.value);
			break;
		}
	}

	/** Chooses a random capacity for the hash tables. */
	@Override
	public void initialize() {
		Random rand = new Random();
		nrBuckets = rand.nextInt(MAX_NRBUCKETS) + 1;
	}

	/** Sets up the hash tables with the desired capacity. */
	@Override
	public void setUp() {
		setObjects(new LLHashTable(nrBuckets), new LLHashTable(nrBuckets));
	}

	@Override
	public LLHashTableAction[] randomActions(int n) {
		Random rand = new Random();
		LLHashTableAction res[] = new LLHashTableAction[n];
		for (int i = 0; i < n; i++) {
			switch (rand.nextInt(4)) {
			case 0:
				res[i] = new LLHashTableAction(
						LLHashTableAction.ActionType.PUT,
						rand.nextInt(MAX_KEY), rand.nextInt());
				break;
			case 1:
				res[i] = new LLHashTableAction(
						LLHashTableAction.ActionType.REMOVE,
						rand.nextInt(MAX_KEY), 0);
				break;
			case 2:
				res[i] = new LLHashTableAction(
						LLHashTableAction.ActionType.GET,
						rand.nextInt(MAX_KEY), 0);
				break;
			case 3:
				res[i] = new LLHashTableAction(
						LLHashTableAction.ActionType.OUTPUT, 0, rand.nextInt());
				break;
			}
		}
		return res;
	}

	/**
	 * For a failing test case, tries to minimize the capacity of the hash table
	 * (while keeping the test failing).
	 */
	public void minimizeCapacity() {
		while (nrBuckets > 0 && precondition() && !runTest())
			nrBuckets--;
		nrBuckets = nrBuckets + 1;
	}

	public void printTestCode() {
		System.out.println("  Code: ");
		System.out.println("HashTable h1 = new HashTable(" + nrBuckets + ");");
		System.out
				.println("HashTable h2 = new HashTable(" + nrBuckets + ");\n");

		if (currentTest.withPrefix())
			for (Object action : this.commonPrefix)
				printAction("h1", (LLHashTableAction) action);
		for (Object action : this.command1)
			printAction("h1", (LLHashTableAction) action);
		if (currentTest.withSuffix())
			for (Object action : this.commonSuffix)
				printAction("h1", (LLHashTableAction) action);
		System.out.println("");

		if (currentTest.withPrefix())
			for (Object action : this.commonPrefix)
				printAction("h2", (LLHashTableAction) action);
		for (Object action : this.command2)
			printAction("h2", (LLHashTableAction) action);
		if (currentTest.withSuffix())
			for (Object action : this.commonSuffix)
				printAction("h2", (LLHashTableAction) action);

		System.out.println("\nSystem.out.println(\"\\n\" + h1.equals(h2));");
	}

	private void printAction(String string, LLHashTableAction action) {
		if (action.type == ActionType.OUTPUT) {
			System.out.println("System.out.print(\"" + action.value + " \");");
		} else if (action.type == ActionType.GET) {
			System.out.println("System.out.print(" + string + "." + action
					+ " + \" \");");
		} else {
			System.out.println(string + "." + action + ";");
		}
	}

}
